function [u, v, pu, pv] = HS_flow(I1, I2, u, v, lambda, warps)

[M N C] = size(I1);

[nabla] = make_derivatives(M,N);
laplace = nabla'*nabla;

u = u(:);
v = v(:);

for j = 1:warps
  
  fprintf('HS-flow: warp=%d\n', j);
  
  u0 = u;
  v0 = v;
  
  %% perform warping
  [Ix, Iy, It, I2_warped] = warping(I1, I2, reshape(u0, M, N), ...
    reshape(v0, M, N));
  Ix = Ix(:);
  Iy = Iy(:);
  It = It(:);
  
  a = spdiags(lambda*Ix.^2, 0, M*N, M*N);
  b = spdiags(lambda*Iy.^2, 0, M*N, M*N);
  c = spdiags(lambda*Ix.*Iy, 0, M*N, M*N);
  tmp = Ix.*u0 + Iy.*v0 - It;
  
  %% Form matrix and rhs
  A = [(laplace+a) c; c (laplace+b)];
  rhs = lambda*[Ix.*tmp; Iy.*tmp];
  
  %% solve linear system
  x = A\rhs;
  u = x(1:M*N);
  v = x(M*N+1:end);
  
  show_flow(I1, I2_warped, reshape(u,[M N]),reshape(v, [M N]));
end
  
u = reshape(u, [M N]);
v = reshape(v, [M N]);