function [u, v] = LK_flow(I1, I2, u, v, ws, warps)

[M N C] = size(I1);

af = fspecial('average', [ws ws]);

for j = 1:warps
  
  fprintf('LK-Flow: warp=%d\n', j);
  
  u0 = u;
  v0 = v;
 
  %% Warp image
  idx = repmat([1:N], M,1) + u0;
  idy = repmat([1:M]',1,N) + v0;
  boundary = (idx > N) | (idx < 1) | (idy > M) | (idy < 1);
  
  idx = max(1, min(N, idx));
  idy = max(1, min(M, idy));
  I2_warped = interp2(I2,idx,idy,'bilinear',0);

  %% Compute temproal derivative
  It = I2_warped  - I1;

  %% Compute spatial derivatives
  mask = [-1 0 1]/2;
  Ix = imfilter(I2_warped, mask, 'symmetric');
  Iy = imfilter(I2_warped, mask', 'symmetric');
  
  %% boundary handling
  It(boundary) = 0;
  Ix(boundary) = 0;
  Iy(boundary) = 0;
  
  %% form the elements of the linear system
  a = Ix.^2;
  b = Iy.^2;
  c = Ix.*Iy;
  
  tmp = u0.*Ix +  v0.*Iy - It;
  fu = Ix.*tmp;
  fv = Iy.*tmp;
  
  a = imfilter(a, af, 'symmetric');
  b = imfilter(b, af, 'symmetric');
  c = imfilter(c, af, 'symmetric');
  fu = imfilter(fu, af, 'symmetric');
  fv = imfilter(fv, af, 'symmetric');
  
  det = max(1e-30, a.*b-c.^2);
  u = (b.*fu - c.*fv)./det;
  v = (a.*fv - c.*fu)./det;
  
  show_flow(I1, I2_warped, u,v);
end
