function [u, v, pu, pv] = TVL1_flow(I1, I2, u, v, lambda, epsilon,warps)

[M N C] = size(I1);

[nabla] = make_derivatives(M,N);
nabla_t = nabla';

L = 8/epsilon;

maxiter=100;

u = u(:);
v = v(:);

for j = 1:warps
  
  fprintf('TVL1-flow: warp=%d\n', j);
  
  u0 = u;
  v0 = v;
  
  %% perform warping
  [Ix, Iy, It, I2_warped] = warping(I1, I2, reshape(u0, M, N), ...
    reshape(v0, M, N));
  Ix = Ix(:);
  Iy = Iy(:);
  It = It(:);
  I_grad_sqr = Ix.^2 + Iy.^2 + 1e-09;
  
  % perform FISTA
  tk = 1.0;
  u_ = u;
  v_ = v;
  for iter=1:maxiter
   
    u_old = u;
    v_old = v;
  
    grad_u = nabla*u_;
    grad_v = nabla*v_;
    
    norm_uv = sqrt(grad_u(1:M*N).^2 + grad_u(M*N+1:end).^2 + ...
      grad_v(1:M*N).^2 + grad_v(M*N+1:end).^2);
    
    denom = max(epsilon, norm_uv);
    denom = [denom; denom];
    
    grad_u = nabla_t*(grad_u./denom);
    grad_v = nabla_t*(grad_v./denom);
  
    %% forward step
    u = u_ - grad_u/L;
    v = v_ - grad_v/L;
    
    %% backward step
    %% 2. Proximal l1-minimization
    r = It + (u-u0).*Ix + (v-v0).*Iy;
    idx1 = r < - lambda/L*(I_grad_sqr);
    idx2 = r > + lambda/L*(I_grad_sqr);
    idx3 = abs(r) <= lambda/L*(I_grad_sqr);
    
    u(idx1) = u(idx1) + lambda/L*Ix(idx1);
    u(idx2) = u(idx2) - lambda/L*Ix(idx2);
    u(idx3) = u(idx3) - r(idx3)./I_grad_sqr(idx3).*Ix(idx3);
    
    v(idx1) = v(idx1) + lambda/L*Iy(idx1);
    v(idx2) = v(idx2) - lambda/L*Iy(idx2);
    v(idx3) = v(idx3) - r(idx3)./I_grad_sqr(idx3).*Iy(idx3);
   
    % compute extra point
    tkp=(1+sqrt(1+4*tk^2))/2;
    u_ = u + (tk-1)/(tkp)*(u-u_old);
    v_ = v + (tk-1)/(tkp)*(v-v_old);
    tk = tkp;
    
  end
  
  show_flow(I1, I2_warped, reshape(u,[M N]),reshape(v, [M N]));
end
  
u = reshape(u, [M N]);
v = reshape(v, [M N]);